# pylint: disable=invalid-name, missing-docstring, too-few-public-methods
"""

    This is a straight-forward record augmenting log filter.

    The `gunicorn.access` logic is driven by an oddly delimited
    gunicorn access-log format string

    The access log format.

    ===========  ===========
    Identifier   Description
    ===========  ===========
    h            remote address
    l            ``'-'``
    u            user name
    t            date of the request
    r            status line (e.g. ``GET / HTTP/1.1``)
    m            request method
    U            URL path without query string
    q            query string
    H            protocol
    s            status
    B            response length
    b            response length or ``'-'`` (CLF format)
    f            referer
    a            user agent
    T            request time in seconds
    D            request time in microseconds
    L            request time in decimal seconds
    p            process ID
    {Header}i    request header
    {Header}o    response header
    {Variable}e  environment variable
    ===========  ===========

`!` is '='
'|' delimits the attribute
'remote!%(h)s|method!%(m)s|url-path!%(U)s|query!%(q)s|username!%(u)s|...`

"""
import logging
from .. import LOG_ADDS


class KubeMetaInject(logging.Filter):
    def filter(self, record):
        for k, v in LOG_ADDS.items():
            setattr(record, k, v)

        if record.name == "gunicorn.access":
            msg = record.getMessage()
            try:
                msg_elements = {
                    k: v for k, v in [el.split("!") for el in msg.split("|")]
                }
                record.access = msg_elements
                record.msg = "(access record)"
            except ValueError:
                pass

        return True




# pylint: disable=invalid-name
"""

This is copied directly from `gunicorn.glogging.Logger` class.

    * `format_str` reflects the string generated by jslog4kube
    * the logging system is initialized with the configuration dict
      defined in `jslog4kube.kube.log_config` rather than with
      the default fileConfig mechanism


"""

import logging
from logging.config import dictConfig
import os
import sys
from gunicorn.glogging import Logger
from ..kube.log_config import LOGGING
from .. import format_str


class GunicornLogger(Logger):
    """
    overrides gunicorn's Logger class to provide dictionary configuration
    capabilities
    """

    error_fmt = format_str
    access_fmt = format_str

    def setup(self, cfg):
        self.loglevel = self.LOG_LEVELS.get(cfg.loglevel.lower(), logging.INFO)
        self.error_log.setLevel(self.loglevel)
        self.access_log.setLevel(logging.INFO)

        # set gunicorn.error handler
        if self.cfg.capture_output and cfg.errorlog != "-":
            for stream in sys.stdout, sys.stderr:
                stream.flush()

            self.logfile = open(cfg.errorlog, "a+")
            os.dup2(self.logfile.fileno(), sys.stdout.fileno())
            os.dup2(self.logfile.fileno(), sys.stderr.fileno())

        self._set_handler(
            self.error_log,
            cfg.errorlog,
            logging.Formatter(self.error_fmt, self.datefmt),
        )

        # set gunicorn.access handler
        if cfg.accesslog is not None:
            self._set_handler(
                self.access_log,
                cfg.accesslog,
                fmt=logging.Formatter(self.access_fmt),
                stream=sys.stdout,
            )

        # set syslog handler
        if cfg.syslog:
            self._set_syslog_handler(self.error_log, cfg, self.syslog_fmt, "error")
            self._set_syslog_handler(self.access_log, cfg, self.syslog_fmt, "access")

        dictConfig(LOGGING)



# pylint: disable=invalid-name, missing-docstring, too-few-public-methods
"""

    The environment variables that drive this are `KUBE_META` and
    `KUBE_META_ENV_PREFIX`

    `KUBE_META` is the directory that Kubernetes metadata mountpoints
    can be found.  The default is `/etc/meta`

    `KUBE_META_ENV_PREFIX` specifies the shell variable prefix
    (defaults to `X`) that indicate a variable that should be injected
    into the log record.

    e.g.:
        X_POD_IP=100.96.1.11
        X_NODE_NAME=ip-10-70...


"""

import os

LOG_ADDS = {}

META_PATH = os.environ.get("KUBE_META", None) or "/etc/meta/"

META_ENV_PREFIX = os.environ.get("KUBE_META_ENV_PREFIX", None) or "X"

META_ENV_VAR = {
    k.lower(): v.replace('"', "")
    for k, v in os.environ.items()
    if k.startswith(META_ENV_PREFIX + "_")
}

LOG_ADDS.update(META_ENV_VAR)

try:
    with open(META_PATH + "/annotations") as fh:
        ANNO = {
            k.lower(): v.strip().replace('"', "")
            for k, v in [el.split("=") for el in fh.readlines()]
            if not k.startswith("kubernetes.io")
        }
    LOG_ADDS.update(ANNO)
except IOError:
    pass

try:
    with open(META_PATH + "/labels") as fh:
        LABELS = {
            k.lower(): v.strip().replace('"', "")
            for k, v in [el.split("=") for el in fh.readlines()]
        }
    LOG_ADDS.update(LABELS)
except IOError:
    pass

default_fields = (
    "asctime",
    "message",
    "name",
    "created",
    "filename",
    "module",
    "funcName",
    "lineno",
    "msecs",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "thread",
    "threadName",
    "levelname",
)


format_str = " ".join(
    ["%({})s".format(el) for el in default_fields + tuple(LOG_ADDS.keys())]
)



# pylint: disable=missing-docstring
"""

This is the default logging config dictionary.

Of note:

  * the format string (`format_str`) is a a space-delimited series of
    python logging attributes that include the attributes derived from
    the kubernetes environment and/or meta-data files.

    TODO:
        KubeMetaInject: support exclusion and inclusion arguments
        Since `format_str` is derived from a list of built-in and
        kube-derived keys, manipulating this list before it is applied
        to the log record is a good place for this sort of thing.

        In addition, having `format_str` always reflect the list
        of desired attributes will avoid having to deal with manual
        maintenance of said format string.

    (https://docs.python.org/3.5/library/logging.html#logrecord-attributes)

  * If `gunicorn` is not in play, the `gunicorn.` loggers and the `json-access`
    formatter can be left off
"""

from .. import format_str

DATEFORMAT = "%Y-%m-%dT%H:%M:%S,%03d"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {"format": "%(message)s"},
        "json": {
            "format": format_str,
            "datefmt": DATEFORMAT,
            "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
        },
        "json-access": {
            "datefmt": DATEFORMAT,
            "format": format_str + "%(access)",
            "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
        },
    },
    "filters": {"default": {"()": "jslog4kube.KubeMetaInject"}},
    "handlers": {
        "json-stdout": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "formatter": "json",
            "filters": ["default"],
        },
        "default": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "formatter": "default",
            "filters": ["default"],
        },
    },
    "loggers": {
        "efk": {
            "handlers": ["json-stdout"],
            "propagate": True,
            "level": "INFO",
            "filters": ["default"],
            "formatters": ["json"],
        },
        "demo": {
            "handlers": ["json-stdout"],
            "propagate": True,
            "level": "INFO",
            "filters": ["default"],
            "formatters": ["json"],
        },
        "django": {
            "handlers": ["json-stdout"],
            "propagate": True,
            "level": "INFO",
            "filters": ["default"],
            "formatters": ["json"],
        },
        "gunicorn": {
            "handlers": ["json-stdout"],
            "formatters": ["json"],
            "propagate": False,
            "level": "ERROR",
        },
        "gunicorn.access": {
            "handlers": ["json-stdout"],
            "formatters": ["json-access"],
            "propagate": False,
            "level": "INFO",
        },
        "gunicorn.error": {
            "handlers": ["json-stdout"],
            "formatters": ["json"],
            "propagate": False,
            "level": "INFO",
        },
        "requests": {
            "handlers": ["json-stdout"],
            "formatters": ["json"],
            "propagate": True,
            "filters": ["default"],
            "level": "INFO",
        },
    },
}



from setuptools import setup

setup(
    name="jslog4kube",
    version="1.0.4",
    description="relatively hassle-free JSON logging for Kubernetes pods",
    url="http://gitlab.com/stephen6/jslog4kube",
    author="Stephen D. Spencer",
    author_email="stephen@revsys.com",
    license="MIT",
    packages=["jslog4kube", "jslog4kube.kube", "jslog4kube.gunicorn"],
    install_requires=["python-json-logger"],
    zip_safe=False,
)



# pylint: disable=invalid-name, missing-docstring, too-few-public-methods

import time
from pythonjsonlogger.jsonlogger import JsonFormatter
from .bootstrap import LOG_ADDS, format_str
from .kube.log_config import LOGGING

JsonFormatter.converter = time.gmtime


try:
    from .gunicorn.dictconfig_logger import GunicornLogger

    HAS_GUNICORN = True
except ImportError:
    HAS_GUNICORN = False


__all__ = ["LOG_ADDS", "format_str", "LOGGING"]
